"""
Tests for the tags module.
"""

import unittest

from kerko.tags import TagGate


class TagGateTestCase(unittest.TestCase):

    def setUp(self):
        self.objects = [
            {  # 0
                'tags': [{'tag': '_include'}, ]
            },
            {  # 1
                'tags': [
                    {'tag': 'foo'},
                    {'tag': 'bar'},
                    {'tag': '_include'},
                    {'tag': 'baz'},
                    {'tag': 'qux'},
                ]
            },
            {  # 2
                'tags': [{'tag': '_exclude'}, ]
            },
            {  # 3
                'tags': [
                    {'tag': 'foo'},
                    {'tag': 'bar'},
                    {'tag': '_exclude'},
                    {'tag': 'baz'},
                    {'tag': 'qux'},
                ]
            },
            {  # 4
                'tags': [
                    {'tag': '_include'},
                    {'tag': '_exclude'},
                ]
            },
            {  # 5
                'tags': []
            },
            {  # 6
                'tags': [
                    {'tag': 'foo'},
                    {'tag': 'bar'},
                ]
            },
        ]

    def test_default(self):
        gate = TagGate()
        self.assertTrue(gate.check(self.objects[0]))
        self.assertTrue(gate.check(self.objects[1]))
        self.assertTrue(gate.check(self.objects[2]))
        self.assertTrue(gate.check(self.objects[3]))
        self.assertTrue(gate.check(self.objects[4]))
        self.assertTrue(gate.check(self.objects[5]))
        self.assertTrue(gate.check(self.objects[6]))

    def test_match_include(self):
        gate = TagGate(include_re=r'_include')
        self.assertTrue(gate.check(self.objects[0]))
        self.assertTrue(gate.check(self.objects[1]))
        self.assertFalse(gate.check(self.objects[2]))
        self.assertFalse(gate.check(self.objects[3]))
        self.assertTrue(gate.check(self.objects[4]))
        self.assertFalse(gate.check(self.objects[5]))
        self.assertFalse(gate.check(self.objects[6]))

    def test_match_include_re(self):
        gate = TagGate(include_re=r'_incl.*')
        self.assertTrue(gate.check(self.objects[0]))
        self.assertTrue(gate.check(self.objects[1]))
        self.assertFalse(gate.check(self.objects[2]))
        self.assertFalse(gate.check(self.objects[3]))
        self.assertTrue(gate.check(self.objects[4]))
        self.assertFalse(gate.check(self.objects[5]))
        self.assertFalse(gate.check(self.objects[6]))

    def test_match_exclude(self):
        gate = TagGate(exclude_re=r'_exclude')
        self.assertTrue(gate.check(self.objects[0]))
        self.assertTrue(gate.check(self.objects[1]))
        self.assertFalse(gate.check(self.objects[2]))
        self.assertFalse(gate.check(self.objects[3]))
        self.assertFalse(gate.check(self.objects[4]))
        self.assertTrue(gate.check(self.objects[5]))
        self.assertTrue(gate.check(self.objects[6]))

    def test_match_exclude_re(self):
        gate = TagGate(exclude_re=r'_excl.*')
        self.assertTrue(gate.check(self.objects[0]))
        self.assertTrue(gate.check(self.objects[1]))
        self.assertFalse(gate.check(self.objects[2]))
        self.assertFalse(gate.check(self.objects[3]))
        self.assertFalse(gate.check(self.objects[4]))
        self.assertTrue(gate.check(self.objects[5]))
        self.assertTrue(gate.check(self.objects[6]))

    def test_match_both(self):
        gate = TagGate(include_re=r'_include', exclude_re=r'_exclude')
        self.assertTrue(gate.check(self.objects[0]))
        self.assertTrue(gate.check(self.objects[1]))
        self.assertFalse(gate.check(self.objects[2]))
        self.assertFalse(gate.check(self.objects[3]))
        self.assertFalse(gate.check(self.objects[4]))
        self.assertFalse(gate.check(self.objects[5]))
        self.assertFalse(gate.check(self.objects[6]))



"""
Date and time utilities.
"""

from datetime import datetime

from babel import dates
from flask_babel import get_locale, get_timezone


def reformat_date(value, **kwargs):
    """
    Reformat a date-time string into a localized form.

    A strict parsing of the value is performed, based on the ISO 8601 format
    (calendar dates with time and timezone designator are supported, other
    variants are not). If the value cannot be parsed, the original value is
    returned as is. Date values in Zotero are often manually entered by the
    user, lacking parts such as the day, the time, or the timezone; those values
    will be returned as is.

    :param str value: The value to parse.

    :param dict kwargs: Parameters to pass along to `format_datetime`, once the
        value is parsed.

    :return str: The localized version of the date value, if `value` could be
        parsed, otherwise the original `value` is returned as is.
    """

    # Note: Full ISO 8601 support with strptime was not possible before Python
    # 3.7, because the %z directive didn't support the colon separator.
    # Reference: https://docs.python.org/3/library/datetime.html#technical-detail
    formats_to_try = [
        '%Y-%m-%dT%H:%M:%S.%fZ', '%Y-%m-%dT%H:%M:%S.%f%z',
        '%Y-%m-%dT%H:%M:%SZ', '%Y-%m-%dT%H:%M:%S%z',
        '%Y-%m-%dT%H:%MZ', '%Y-%m-%dT%H:%M%z',
        '%Y-%m-%dT%HZ', '%Y-%m-%dT%H%z',
    ]

    for fmt in formats_to_try:
        try:
            parsed = datetime.strptime(value, fmt)
        except ValueError:
            pass
        else:
            return format_datetime(parsed, **kwargs)
    return value


def format_datetime(dt, *, convert_tz=False, show_tz=False):
    """
    Format a `datetime` object into a localized form.

    :param datetime dt: The datetime object to format.

    :param bool convert_tz: Whether to convert the value to Babel's timezone.

    :param bool show_tz: Whether to print the timezone's name in the result.
        This can only be set to `True` when `convert_tz` is also `True`.

    :return str: The localized version of the datetime object.
    """
    if show_tz and not convert_tz:
        # Only implemented showing the timezone when converting.
        raise ValueError

    parts = [
        dates.format_datetime(
            dt, format='short', tzinfo=get_timezone() if convert_tz else None, locale=get_locale()
        ),
    ]
    if show_tz:
        parts.append(f"({get_timezone().tzname(dt)})")
    return ' '.join(parts)



#!/usr/bin/env python

"""
Package build setup script.

Metadata and options are specified in `setup.cfg`.

Reference: https://setuptools.readthedocs.io/en/latest/setuptools.html
"""

import setuptools
from setuptools.command.develop import develop as BaseDevelop
from setuptools.command.sdist import sdist as BaseSDist
try:
    from wheel.bdist_wheel import bdist_wheel as BaseBDistWheel
except ImportError:
    BaseBDistWheel = None


class CompileCatalogMixin():
    """
    Compile MO files with Babel's ``compile_catalog`` command.

    This happens after installing in development mode, or before building sdist
    and wheel.
    """

    # Note: Inspired by WTForms' setup.py.

    def run(self):
        is_develop = isinstance(self, Develop)

        if not is_develop:
            self.run_command('compile_catalog')

        super().run()

        if is_develop and not self.uninstall:
            self.run_command('compile_catalog')


class Develop(CompileCatalogMixin, BaseDevelop):
    pass


class SDist(CompileCatalogMixin, BaseSDist):
    pass


cmd_classes = {
    'develop': Develop,
    'sdist': SDist,
}

if BaseBDistWheel:

    class BDistWheel(CompileCatalogMixin, BaseBDistWheel):
        pass

    cmd_classes['bdist_wheel'] = BDistWheel


if __name__ == "__main__":
    setuptools.setup(cmdclass=cmd_classes)



from werkzeug.datastructures import MultiDict

from flask import current_app, url_for


class Criteria:
    """Represent a complete set of user-submitted search criteria."""

    # This class is close to the views because it knows about all sorts
    # of view-related elements, such as the request and the form fields.

    def __init__(self, request=None):
        """Extract criteria from the request, ensuring that all are valid."""
        self.composer = current_app.config['KERKO_COMPOSER']
        self._extract_keywords(request)
        self._extract_filters(request)
        self._extract_pager(request)
        self._extract_sort(request)  # Must run after _extract_keywords().
        self._extract_show_abstracts(request)
        self._extract_print_preview(request)
        self._extract_id(request)  # Must run after _extract_pager().

    def _extract_keywords(self, request=None):
        self.keywords = MultiDict()
        if request:
            for scope in self.composer.scopes.values():
                if scope.key in request.args:
                    self.keywords.setlist(scope.key, request.args.getlist(scope.key))

    def _extract_filters(self, request=None):
        self.filters = MultiDict()
        if request:
            for spec in self.composer.facets.values():
                values = request.args.getlist(spec.filter_key)
                if values:
                    self.filters.setlist(spec.filter_key, values)

    def _extract_pager(self, request=None):
        if request:
            try:
                self.page_num = int(request.args.get('page', 1))
            except ValueError:
                self.page_num = 1
        else:
            self.page_num = 1
        if request and request.args.get('page-len') == 'all':
            # Note: print_preview pages show all results.
            self.page_len = None
        else:
            default_page_len = current_app.config['KERKO_PAGE_LEN']
            if request:
                try:
                    self.page_len = int(request.args.get('page-len', default_page_len))
                except ValueError:
                    self.page_len = default_page_len
            if not request or self.page_len < 1:
                self.page_len = default_page_len

    def _extract_sort(self, request=None):
        composer = current_app.config['KERKO_COMPOSER']
        # Use the first enabled sort as default.
        default_sort = None
        for sort_spec in composer.get_ordered_specs('sorts'):
            if sort_spec.is_allowed(criteria=self):
                default_sort = sort_spec.key
                break
        assert default_sort

        if request:
            self.sort = request.args.get('sort', None)
        if not request \
                or self.sort not in composer.sorts \
                or not composer.sorts[self.sort].is_allowed(criteria=self):
            self.sort = default_sort

    def _extract_show_abstracts(self, request=None):
        if current_app.config['KERKO_RESULTS_ABSTRACTS_TOGGLER']:
            self.show_abstracts = self.get_boolean_param(
                request, param='abstracts', default=current_app.config['KERKO_RESULTS_ABSTRACTS']
            )
        else:
            self.show_abstracts = current_app.config['KERKO_RESULTS_ABSTRACTS']

    def _extract_print_preview(self, request=None):
        self.print_preview = self.get_boolean_param(request, param='print-preview', default=False)

    def _extract_id(self, request=None):
        if request and self.page_len == 1:
            self.id = request.args.get('id', None)
        else:
            self.id = None

    @staticmethod
    def get_boolean_param(request, param, default):
        if request:
            value = request.args.get(param)
            if value not in ['t', 'f', '0', '1']:
                return default
            return value in ['t', '1']
        return default

    def fit_pager(self, page_count):
        """Ensure that pager values fit within the given page count."""
        self.page_num = min(self.page_num, page_count)

    def has_keyword_search(self):
        """Return True if criteria contains a keyword search."""
        return self.keywords is not None and any(self.keywords.values())

    def has_filter_search(self):
        """Return True if criteria contains active filters."""
        return self.filters is not None and any(self.filters.values())

    def build_query(
            self,
            *,
            keywords=None,
            filters=None,
            page_num=None,
            page_len=None,
            sort=None,
            show_abstracts=None,
            print_preview=None,
            id_=None
    ):
        """
        Prepare a query string ready for use when generating an URL.

        :param MultiDict keywords: New keywords dict. Use None to preserve the
            active keywords, or False to reset.

        :param MultiDict filters: New filters dict. Use None to preserve the
            active filters, or False to reset.

        :param str sort: Key of the sort specification. Use None to preserve the
            active sort, or False to reset.
        """
        query = MultiDict()
        if keywords:
            query.update(keywords)
        elif keywords is None:
            query.update(self.keywords)
        if filters:
            query.update(filters)
        elif filters is None:
            query.update(self.filters)
        if page_num:
            # Never defaults to self.page_num, because new queries should
            # always reset the pager.
            query['page'] = page_num
        if page_len and page_len != current_app.config['KERKO_PAGE_LEN']:
            # Never defaults to self.page_len, because new queries should
            # always reset the pager.
            query['page-len'] = page_len
        if sort:
            query['sort'] = sort
        elif sort is None:
            query['sort'] = self.sort
        if show_abstracts is None:
            show_abstracts = self.show_abstracts
        if show_abstracts != current_app.config['KERKO_RESULTS_ABSTRACTS']:
            query['abstracts'] = int(show_abstracts)
        if print_preview or self.print_preview:
            # Only set print-preview if it is enabled.
            query['print-preview'] = 1
        if id_ and page_len == 1:
            # Never defaults to self.id, because new queries should never lead
            # to the same item.
            query['id'] = id_
        return query.to_dict(flat=False)

    def build_url(self, **kwargs):
        """Build an URL with all the search criteria."""
        return url_for('kerko.search', **self.build_query(**kwargs))

    def build_download_url(self, citation_format_key, **kwargs):
        """Build a citation download URL with all the search criteria."""
        return url_for(
            'kerko.search_citation_download',
            citation_format_key=citation_format_key,
            **self.build_query(**kwargs)
        )

    def build_add_keywords_url(self, scope, value):
        """
        Build an URL with all the search criteria, adding the given keywords.

        :param str scope: Key of the keyword search scope.

        :param str value: Keywords to search the scope with.
        """
        value = value.strip()
        new_keywords = self.keywords.deepcopy()
        new_filters = None
        new_sort = None
        if scope in self.composer.scopes and value:
            if not new_keywords:  # If adding keywords for the first time.
                new_sort = False  # Reset the sort.
            new_keywords.add(scope, value)
        return url_for(
            'kerko.search',
            **self.build_query(keywords=new_keywords, filters=new_filters, sort=new_sort)
        )

    def build_remove_keywords_url(self, scope, value):
        """Build an URL with all the search criteria, except the given keywords."""
        new_keywords = self.keywords.deepcopy()
        if new_keywords:
            new_values = [v for v in new_keywords.poplist(scope) if v != value]
            if new_values:
                new_keywords.setlist(scope, new_values)
        return url_for('kerko.search', **self.build_query(keywords=new_keywords))

    def build_add_filter_url(self, facet_spec, value):
        """Build an URL with all the search criteria, adding the given facet value."""
        new_filters = facet_spec.add_filter(value, self.filters)
        if new_filters:
            return url_for('kerko.search', **self.build_query(filters=new_filters))
        return None

    def build_remove_filter_url(self, facet_spec, value):
        """Build an URL with all the search criteria, removing the given facet value."""
        new_filters = facet_spec.remove_filter(value, self.filters)
        if new_filters:
            return url_for('kerko.search', **self.build_query(filters=new_filters))
        return None



"""Tags-related utilities."""

import re


class TagGate:
    """Determine whether an object is included or excluded based on its tags."""

    def __init__(self, include_re='', exclude_re=''):
        """
        Initialize the instance.

        :param str include_re: Regex to use to include objects based on their
            tags. Any object which does not have a tag that matches this regular
            expression will be excluded. If empty (which is the default), all
            objects will be included unless the `exclude_re` argument is set and
            causes some to be excluded.

        :param str exclude_re: Regex to use to exclude objects based on their
            tags. Any object that have a tag that matches this regular
            expression will be excluded. If empty (which is the default), no
            objects will be excluded unless the `include_re` argument is set, in
            which case items that don't have any tag that matches it will be
            excluded.
        """
        self.include_re = re.compile(include_re) if include_re else None
        self.exclude_re = re.compile(exclude_re) if exclude_re else None

    def check(self, obj):
        """
        Check whether the object is to be included or excluded.

        :param dict obj: A dict which is expected to contain an optional 'tags'
            key, whose corresponding value should be a list of dicts, each with
            a 'tag' key, whose corresponding value should be a string
            representing a tag.
        """
        included = self.include_re is None
        excluded = False
        if self.include_re or self.exclude_re:
            for tag_data in obj.get('tags', []):
                tag = tag_data.get('tag', '').strip()
                if self.include_re and self.include_re.match(tag):
                    included = True
                if self.exclude_re and self.exclude_re.match(tag):
                    excluded = True
        return included and not excluded



from time import sleep
import requests

import wrapt

from flask import current_app
from pyzotero import zotero, zotero_errors


@wrapt.decorator
def retry_zotero(wrapped, _instance, args, kwargs):
    """
    Retry the wrapped function if the Zotero API call fails.

    Caution: This decorator should only be used on simple functions, as the
    whole function is called repeatedly as long a Zotero fails.
    """
    attempts = 1
    while True:
        try:
            return wrapped(*args, **kwargs)
        except (
                requests.exceptions.ConnectionError,
                zotero_errors.HTTPError,
                zotero_errors.UnsupportedParams
        ) as e:
            current_app.logger.exception(e)
            if attempts < current_app.config['KERKO_ZOTERO_MAX_ATTEMPTS']:
                current_app.logger.warning(
                    "The Zotero API call has failed in {func}. "
                    "New attempt in {wait} seconds...".format(
                        func=wrapped.__name__,
                        wait=current_app.config['KERKO_ZOTERO_WAIT']
                    )
                )
                attempts += 1
                sleep(current_app.config['KERKO_ZOTERO_WAIT'])
            else:
                current_app.logger.error(
                    "The maximum number of API call attempts to Zotero has "
                    "been reached. Stopping."
                )
                raise


def init_zotero():
    return zotero.Zotero(
        library_id=current_app.config['KERKO_ZOTERO_LIBRARY_ID'],
        library_type=current_app.config['KERKO_ZOTERO_LIBRARY_TYPE'],
        api_key=current_app.config['KERKO_ZOTERO_API_KEY'],
        locale=current_app.config['KERKO_ZOTERO_LOCALE']
    )


@retry_zotero
def load_item(zotero_credentials, item_key):
    """Return a specific item."""
    return zotero_credentials.item(item_key)


@retry_zotero
def load_item_types(zotero_credentials):
    """Return all Zotero item types."""
    current_app.logger.info("Requesting all item types...")
    return zotero_credentials.item_types()


@retry_zotero
def load_item_fields(zotero_credentials):
    """Return all Zotero fields."""
    current_app.logger.info("Requesting all fields...")
    return zotero_credentials.item_fields()


@retry_zotero
def load_item_type_fields(zotero_credentials, item_type):
    """Return all Zotero fields for a given item type."""
    current_app.logger.info(
        "Requesting fields for items of type '{item_type}'...".format(
            item_type=item_type
        )
    )
    return zotero_credentials.item_type_fields(item_type)


@retry_zotero
def load_item_type_creator_types(zotero_credentials, item_type):
    """List all Zotero creator types for a given item type."""
    current_app.logger.info(
        "Requesting creator types for items of type '{item_type}'...".format(
            item_type=item_type
        )
    )
    return zotero_credentials.item_creator_types(item_type)


@retry_zotero
def retrieve_file(zotero_credentials, item_id):
    """Retrieve the file from the given item."""
    return zotero_credentials.file(item_id)


class Collections:
    """
    Iterable over all collections found in Zotero.

    Collections are loaded in memory all at once, and may be acccessed by key
    using the dict[] notation.

    Each collection is a dict as returned by Zotero. Example::

        {
            'data': {
                'key': 'C435CGN4',
                'name': 'Example Collection Name',
                'parentCollection': 'D5U2TEVB',
                'relations': {   },
                'version': 114
            },
            'key': 'C435CGN4',
            'library': {
                'id': 2076305,
                'links': {
                    'alternate': {
                        'href': 'https://www.example.com',
                        'type': 'text/html'
                    }
                },
                'name': 'Example Library',
                'type': 'user'
            },
            'links': {
                'alternate': {
                    'href': 'https://www.zotero.org/example_library/collections/C435CGN4',
                    'type': 'text/html'
                },
                'self': {
                    'href': 'https://api.zotero.org/users/9999999/collections/C435CGN4',
                    'type': 'application/json'
                },
                'up': {
                    'href': 'https://api.zotero.org/users/9999999/collections/D5U2TEVB',
                    'type': 'application/atom+xml'
                }
            },
            'meta': {
                'numCollections': 0,
                'numItems': 15
            },
            'version': 114
        }
    """

    def __init__(self, zotero_credentials, top_level=False):
        self.collections = {}
        # Immediately load all collections, to allow later access by key.
        self.load_collections(zotero_credentials, top_level)
        self.iterator = None

    @retry_zotero
    def load_collections(self, zotero_credentials, top_level):
        current_app.logger.info(
            "Requesting {which} collections...".format(
                which='top-level' if top_level else 'all'
            )
        )
        start = 0
        if top_level:
            method = zotero_credentials.collections_top
        else:
            method = zotero_credentials.collections
        while True:
            more = method(start=start, limit=current_app.config['KERKO_ZOTERO_BATCH_SIZE'])
            if not more:
                break
            start += len(more)
            for collection in more:
                self.collections[collection['key']] = collection

    def __iter__(self):
        return self

    def __len__(self):
        return len(self.collections)

    def __getitem__(self, key):
        return self.collections[key]

    def __contains__(self, key):
        return key in self.collections

    def __next__(self):
        if self.iterator is None:
            self.iterator = iter(self.collections.values())
        return next(self.iterator)

    def get(self, key, default):
        return self.collections.get(key, default)

    def ancestors(self, key):
        """
        Return the ancestors of the specified collection.

        :return: A list representing the path of the specified collection,
            including the specified collection. Each list element is a
            collection key.
        """
        collection = self.collections.get(key)
        if collection:
            parent_key = collection['data'].get('parentCollection')
            if parent_key:
                ancestors = self.ancestors(parent_key)
                ancestors.append(key)
                return ancestors
            return [key]
        return []


class Items:
    """
    Iterable over Zotero items.

    Items are loaded on demand, in small batches, and cannot be accessed
    directly, but can be iterated on. Each item is represented by a (item,
    children) tuple, where item is a dict as returned by Zotero, and chidren a
    list of the child items (also dicts as returned by Zotero).
    """

    def __init__(self, zotero_credentials, item_types=None, formats=None):
        """
        Construct the iterable.

        :param zotero.Zotero zotero_credentials: The Zotero instance.

        :param Iterable item_types: Iterable of desired Zotero item types. If
            None, all items will be retrieved.

        :param Iterable formats: Iterable of format values for the Zotero read.
            Defaults to `['data']`. See available formats on
            https://www.zotero.org/support/dev/web_api/v3/basics#parameters_for_format_json.
        """
        self.zotero_credentials = zotero_credentials
        self.item_type_filter = ' || '.join(item_types) if item_types else None
        self.include = ','.join(formats or ['data'])
        self.start = current_app.config['KERKO_ZOTERO_START']
        self.zotero_batch = []
        self.iterator = iter(self.zotero_batch)

    def __iter__(self):
        return self

    def __next__(self):
        if current_app.config['KERKO_ZOTERO_END'] > 0 and \
                self.start >= current_app.config['KERKO_ZOTERO_END']:
            raise StopIteration
        while True:
            try:
                return self._next_item()
            except StopIteration:
                self._next_batch()

    @retry_zotero
    def _next_batch(self):
        limit = current_app.config['KERKO_ZOTERO_BATCH_SIZE']
        current_app.logger.info(
            "Requesting up to {limit} items starting at position {start}...".format(
                limit=limit, start=self.start
            )
        )
        self.zotero_batch = self.zotero_credentials.items(
            start=self.start,
            limit=limit,
            sort='dateModified',
            direction='asc',
            include=self.include,
            style=current_app.config['KERKO_CSL_STYLE'],
            itemType=self.item_type_filter
        )
        if not self.zotero_batch:
            raise StopIteration  # Empty batch, nothing more to iterate on.
        self.iterator = iter(self.zotero_batch)

    def _next_item(self):
        zotero_item = next(self.iterator)
        self.start += 1
        return zotero_item


class ChildItems:
    """
    Iterable over Zotero child items.

    Children are loaded on demand, in small batches, and cannot be accessed
    directly, but can be iterated on. Each child is represented by a dict as
    returned by Zotero.
    """

    def __init__(self, zotero_credentials, item_key, item_types=None, formats=None):
        self.zotero_credentials = zotero_credentials
        self.item_key = item_key
        self.item_type_filter = ' || '.join(item_types) if item_types else None
        self.include = ','.join(formats or ['data'])
        self.start = 0
        self.zotero_batch = []
        self.iterator = iter(self.zotero_batch)

    def __iter__(self):
        return self

    def __next__(self):
        while True:
            try:
                return self._next_item()
            except StopIteration:
                self._next_batch()

    @retry_zotero
    def _next_batch(self):
        limit = current_app.config['KERKO_ZOTERO_BATCH_SIZE']
        if self.zotero_batch and len(self.zotero_batch) < limit:
            # Previous batch did not reach limit. No need for more batches.
            raise StopIteration
        current_app.logger.info(
            "Requesting up to {limit} children for item {item_key}"
            " starting at position {start}...".format(
                limit=limit, item_key=self.item_key, start=self.start
            )
        )
        self.zotero_batch = self.zotero_credentials.children(
            self.item_key,
            start=self.start,
            limit=limit,
            sort='dateModified',
            direction='asc',
            include=self.include,
            itemType=self.item_type_filter,
        )
        if not self.zotero_batch:
            raise StopIteration  # Empty batch, nothing more to iterate.
        self.iterator = iter(self.zotero_batch)

    def _next_item(self):
        zotero_item = next(self.iterator)
        self.start += 1
        return zotero_item
